"""Cloud protocols"""

import asyncio
import logging
from abc import abstractmethod
from pathlib import Path
from typing import Callable, Mapping, Optional, Sequence, TypeVar, Union

from asyncssh.public_key import SSHKey
from attr import define
from typing_extensions import Protocol, Self

from ..config import ConfigCloud, ConfigLocal, ConfigRemote, EngineRepository
from ..db import DB

SupportedPlatformChecker = Callable[[str], bool]

TConfigCloud_contra = TypeVar(
    "TConfigCloud_contra", bound=ConfigCloud, contravariant=True
)


class PCloudConfig(Protocol):
    "Cloud config init protocol"

    bootcmd: Sequence[Union[str, Sequence[str]]]
    package_upgrade: bool
    packages: Sequence[str]

    @abstractmethod
    def render(self) -> str:
        "Render config to string"
        raise NotImplementedError

    @abstractmethod
    def render_base64(self) -> str:
        "Render to user-data format as base64 string"


class CreateNodeCallable(Protocol[TConfigCloud_contra]):
    "Create node in the cloud protocol"

    @abstractmethod
    async def __call__(
        self,
        log: logging.Logger,
        cfg: TConfigCloud_contra,
        key: SSHKey,
        cloud_config: Optional[PCloudConfig] = None,
    ) -> str:
        raise NotImplementedError


class DeleteNodeCallable(Protocol[TConfigCloud_contra]):
    "Delete node in the cloud protocol"

    @abstractmethod
    async def __call__(
        self,
        log: logging.Logger,
        cfg: TConfigCloud_contra,
        host: str,
    ) -> None:
        raise NotImplementedError


class PCloudAdapter(Protocol[TConfigCloud_contra]):
    "Cloud adapter protocol"

    name: str
    supported_platform_checks: Sequence[SupportedPlatformChecker]
    create_node: CreateNodeCallable[TConfigCloud_contra]
    create_node_conn_timeout: int
    create_node_timeout: int
    delete_node: DeleteNodeCallable[TConfigCloud_contra]
    op_limit: int

    @classmethod
    @abstractmethod
    def create(
        cls,
        name: str,
        supported_platform_checks: Sequence[SupportedPlatformChecker],
        create_node: CreateNodeCallable[TConfigCloud_contra],
        delete_node: DeleteNodeCallable[TConfigCloud_contra],
        create_node_conn_timeout: Optional[int],
        create_node_timeout: Optional[int],
        op_limit: int = 1,
    ) -> Self:
        "Create adapter"
        raise NotImplementedError

    @abstractmethod
    def get_op_semaphore(self) -> asyncio.Semaphore:
        """
        Cached semaphore getter.
        It's because you cannot create async semaphore outside the loop.
        "attached to a different loop" error.
        """
        raise NotImplementedError


class PCloudAPI(Protocol[TConfigCloud_contra]):
    "Cloud API protocol"

    name: str
    config: TConfigCloud_contra
    local_config: ConfigLocal
    remote_config: ConfigRemote
    engines: EngineRepository
    log: logging.Logger

    @classmethod
    @abstractmethod
    async def create(
        cls,
        adapter: PCloudAdapter[TConfigCloud_contra],
        config: TConfigCloud_contra,
        local_config: ConfigLocal,
        remote_config: ConfigRemote,
        engines: EngineRepository,
        ssh_key_lock: Optional[asyncio.Lock] = None,
        log: Optional[logging.Logger] = None,
    ) -> Self:
        "Create cloud API"
        raise NotImplementedError

    @abstractmethod
    def get_op_semaphore(self) -> asyncio.Semaphore:
        """
        Cached semaphore getter.
        It's because you cannot create async semaphore outside the loop.
        "attached to a different loop" error.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_cloud_config_data(self) -> PCloudConfig:
        "Return cloud config init"
        raise NotImplementedError

    @abstractmethod
    def is_platform_supported(self, platform: str) -> bool:
        "Is platform is supported by cloud?"
        raise NotImplementedError

    @abstractmethod
    async def create_node(self) -> str:
        "Create new node"
        raise NotImplementedError

    @abstractmethod
    async def delete_node(self, host: str):
        "Delete node"
        raise NotImplementedError


@define(frozen=True)
class CloudCapacity:
    "Cloud capacity object"

    name: str
    max: int
    current: int


class PCloudAPIManager(Protocol):
    "Cloud API manager protocol"

    apis: Mapping[str, PCloudAPI]
    db: DB
    log: logging.Logger
    keys_dir: Path

    @classmethod
    @abstractmethod
    async def create(
        cls,
        db: DB,
        local_config: ConfigLocal,
        remote_config: ConfigRemote,
        cloud_configs: Sequence[ConfigCloud],
        engines: EngineRepository,
        log: Optional[logging.Logger] = None,
    ) -> Self:
        "Create cloud API manager"
        raise NotImplementedError

    @abstractmethod
    def __bool__(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def stop(self) -> None:
        "Stop cloud api manager"
        raise NotImplementedError

    @abstractmethod
    def mark_task_done(self, on_task: int) -> None:
        "Mark the task for which the creation of a node was requested as completed"
        raise NotImplementedError

    @abstractmethod
    async def get_capacity(self) -> Mapping[str, CloudCapacity]:
        "Get clouds capacity"
        raise NotImplementedError

    @abstractmethod
    async def allocate(
        self,
        on_task: Optional[int] = None,
        want_platforms: Optional[Sequence[str]] = None,
        throttle: bool = True,
    ) -> Union[str, None]:
        "Allocate new node"
        raise NotImplementedError

    @abstractmethod
    async def deallocate(self, ip_addr: str):
        "Deallocate the node"
        raise NotImplementedError



"""Async queue with message deduplication"""

import asyncio
from typing import Deque, Generic, Hashable, Set, TypeVar

from attrs import define, field

TUMsgId = TypeVar("TUMsgId", bound=Hashable)
TUMsgPayload = TypeVar("TUMsgPayload")


@define(frozen=True)
class UMessage(Generic[TUMsgId, TUMsgPayload]):
    """Async queue message"""

    id: TUMsgId = field()
    payload: TUMsgPayload = field(hash=False)


class UniqueQueue(asyncio.Queue, Generic[TUMsgId, TUMsgPayload]):
    """Async queue with message deduplication"""

    name: str
    _queue: Deque[UMessage[TUMsgId, TUMsgPayload]]
    _done_pending: Set[UMessage[TUMsgId, TUMsgPayload]]

    def __init__(self, name: str, *argv, maxsize: int = 0, **kwargs):
        self.name = name
        self._done_pending = set()
        super().__init__(maxsize, *argv, **kwargs)

    def _get(self):
        item = self._queue.popleft()
        self._done_pending.add(item)
        return item

    async def get(self) -> UMessage[TUMsgId, TUMsgPayload]:
        return await super().get()

    async def put(self, item: UMessage[TUMsgId, TUMsgPayload]) -> None:
        # skip already added
        if item in self._queue or item in self._done_pending:
            return
        await super().put(item)

    def task_done(self):
        raise NotImplementedError("task_done() not implemented, use item_done()")

    def item_done(self, item: UMessage):
        """Indicate that a enqueued task is complete."""
        self._done_pending.remove(item)
        super().task_done()

    def psize(self):
        """Number of items not done but not in queue."""
        return len(self._done_pending)



"""Azure cloud methods"""

import logging
from pathlib import PurePosixPath
from typing import Dict, Optional, Tuple, cast

from asyncssh.public_key import SSHKey
from attrs import asdict, evolve
from azure.core.credentials_async import AsyncTokenCredential
from azure.core.exceptions import (
    AzureError,
    IncompleteReadError,
    ServiceRequestTimeoutError,
    ServiceResponseError,
    ServiceResponseTimeoutError,
)
from azure.identity.aio import ClientSecretCredential
from azure.mgmt.compute.v2021_07_01.aio import ComputeManagementClient
from azure.mgmt.compute.v2021_07_01.models import (
    BootDiagnostics,
    DiagnosticsProfile,
    DiskCreateOptionTypes,
    DiskDeleteOptionTypes,
    HardwareProfile,
    ImageReference,
    LinuxConfiguration,
    NetworkProfile,
    OSDisk,
    OSProfile,
    SshConfiguration,
    SshPublicKey,
    StorageProfile,
    VirtualMachine,
)
from azure.mgmt.network.v2020_06_01.aio import NetworkManagementClient
from azure.mgmt.network.v2020_06_01.models import (
    IPAllocationMethod,
    NetworkInterface,
    NetworkInterfaceIPConfiguration,
    TagsObject,
)

from ..config.cloud import AzureImageReference, ConfigCloudAzure
from .protocols import PCloudConfig
from .utils import get_rnd_name

# Azure SDK is too noisy
for logger_name in [
    "azure.core.pipeline.policies.http_logging_policy",
    "azure.identity.aio._internal.get_token_mixin",
    "msrest.serialization",
]:
    logging.getLogger(logger_name).setLevel(logging.ERROR)


ID_TAG_NAME = "yascheduler_ip"

RETRY_AZURE_ERRORS = (
    ServiceResponseError,
    ServiceRequestTimeoutError,
    ServiceResponseTimeoutError,
    IncompleteReadError,
)
ALL_AZURE_ERRORS = (AzureError,)


async def create_nic(
    log: logging.Logger,
    cfg: ConfigCloudAzure,
    client: NetworkManagementClient,
    vm_name: str,
) -> Tuple[NetworkInterface, str]:
    "Create network interface"
    nic_name = f"{vm_name}-nic"
    ip_config_name = f"{nic_name}-ip-config"
    subnet = await client.subnets.get(
        resource_group_name=cfg.resource_group,
        virtual_network_name=cfg.vnet,
        subnet_name=cfg.subnet,
    )
    log.debug(f"Subnet {subnet.name} found")
    nsg = await client.network_security_groups.get(cfg.resource_group, cfg.nsg)
    log.debug(f"Network security group {nsg.name} found")
    nic_ip_config_params = NetworkInterfaceIPConfiguration(
        name=ip_config_name,
        subnet=subnet,
        private_ip_allocation_method=IPAllocationMethod.DYNAMIC,
    )
    nic_params = NetworkInterface(
        name=nic_name,
        location=cfg.location,
        ip_configurations=[nic_ip_config_params],
        network_security_group=nsg,
    )
    poller = await client.network_interfaces.begin_create_or_update(
        resource_group_name=cfg.resource_group,
        network_interface_name=nic_name,
        parameters=nic_params,
    )
    await poller.wait()
    nic = await poller.result()
    log.debug(f"Network interface {nic.name} created")
    ip_addr = None
    if nic.ip_configurations:
        for ip_conf in nic.ip_configurations:
            ip_addr = ip_conf.private_ip_address
    if not ip_addr:
        raise RuntimeError("Azure VM created but no IP is assigned")
    await client.network_interfaces.update_tags(
        cfg.resource_group,
        cast(str, nic.name),
        parameters=TagsObject(tags={ID_TAG_NAME: ip_addr}),
    )
    return nic, ip_addr


def create_vm_params(
    location: str,
    vm_name,
    vm_image: AzureImageReference,
    vm_size: str,
    nic: NetworkInterface,
    username: str,
    ssh_key: SSHKey,
    tags: Dict[str, str],
    cloud_config: Optional[PCloudConfig] = None,
) -> VirtualMachine:
    """Create VirtualMachine params"""
    img_ref = ImageReference.from_dict(asdict(vm_image))
    pub_key = SshPublicKey(
        path=str(PurePosixPath("/home", username, ".ssh/authorized_keys")),
        key_data=ssh_key.export_public_key("openssh").decode("utf-8"),
    )
    custom_data = None
    if cloud_config:
        my_boot_cmds = [
            # see https://github.com/MicrosoftDocs/azure-docs/issues/82500
            "systemctl mask waagent-apt.service",
        ]
        custom_data = evolve(
            cloud_config, bootcmd=[*my_boot_cmds, *cloud_config.bootcmd]
        ).render_base64()

    return VirtualMachine(
        location=location,
        tags=tags,
        hardware_profile=HardwareProfile(vm_size=vm_size),
        storage_profile=StorageProfile(
            image_reference=img_ref,
            os_disk=OSDisk(
                create_option=DiskCreateOptionTypes.FROM_IMAGE,
                delete_option=DiskDeleteOptionTypes.DELETE,
            ),
        ),
        network_profile=NetworkProfile(network_interfaces=[nic]),
        os_profile=OSProfile(
            computer_name=vm_name[:15],  # max length 15
            admin_username=username,
            custom_data=custom_data,
            linux_configuration=LinuxConfiguration(
                disable_password_authentication=True,
                ssh=SshConfiguration(public_keys=[pub_key]),
            ),
        ),
        diagnostics_profile=DiagnosticsProfile(
            boot_diagnostics=BootDiagnostics(enabled=True)
        ),
    )


async def create_node(
    nmc: NetworkManagementClient,
    cmc: ComputeManagementClient,
    log: logging.Logger,
    cfg: ConfigCloudAzure,
    key: SSHKey,
    cloud_config: Optional[PCloudConfig] = None,
):
    """Create virtual machine with nic"""
    vm_name = get_rnd_name("yascheduler-vm")
    nic, ip_addr = await create_nic(log=log, cfg=cfg, client=nmc, vm_name=vm_name)
    vm_params = create_vm_params(
        location=cfg.location,
        vm_name=vm_name,
        vm_image=cfg.vm_image,
        vm_size=cfg.vm_size,
        nic=nic,
        username=cfg.username,
        ssh_key=key,
        tags={ID_TAG_NAME: ip_addr},
        cloud_config=cloud_config,
    )

    poller = await cmc.virtual_machines.begin_create_or_update(
        resource_group_name=cfg.resource_group,
        vm_name=get_rnd_name("yascheduler-vm"),
        parameters=vm_params,
    )
    await poller.wait()
    vm_res = await poller.result()
    log.debug(f"VM {vm_res.name} created")
    return ip_addr


async def az_create_node(
    log: logging.Logger,
    cfg: ConfigCloudAzure,
    key: SSHKey,
    cloud_config: Optional[PCloudConfig] = None,
) -> str:
    """Create virtual machine with network interface"""
    async with ClientSecretCredential(
        cfg.tenant_id, cfg.client_id, cfg.client_secret
    ) as cred:
        cred = cast(AsyncTokenCredential, cred)  # fix library type errors
        async with NetworkManagementClient(cred, cfg.subscription_id) as nmc:
            async with ComputeManagementClient(cred, cfg.subscription_id) as cmc:
                return await create_node(nmc, cmc, log, cfg, key, cloud_config)


async def delete_node(
    nmc: NetworkManagementClient,
    cmc: ComputeManagementClient,
    log: logging.Logger,
    cfg: ConfigCloudAzure,
    host: str,
):
    """Delete virtual machine with network interface"""
    async for result in cmc.virtual_machines.list(cfg.resource_group):
        vm_res = cast(VirtualMachine, result)
        tag_ip = (vm_res.tags or {}).get(ID_TAG_NAME)
        if tag_ip == host:
            poller = await cmc.virtual_machines.begin_power_off(
                cfg.resource_group, cast(str, vm_res.name)
            )
            await poller.wait()

            poller = await cmc.virtual_machines.begin_delete(
                cfg.resource_group, cast(str, vm_res.name)
            )
            await poller.wait()
            log.debug(f"VM {vm_res.name} deleted")
            break

    nic = None
    async for result in nmc.network_interfaces.list(cfg.resource_group):
        nic = cast(NetworkInterface, result)
        tag_ip = (nic.tags or {}).get(ID_TAG_NAME)
        if tag_ip == host:
            poller = await nmc.network_interfaces.begin_delete(
                cfg.resource_group, cast(str, nic.name)
            )
            await poller.wait()
            log.debug(f"Network interface {nic.name} deleted")
            break


async def az_delete_node(
    log: logging.Logger,
    cfg: ConfigCloudAzure,
    host: str,
) -> None:
    """Delete virtual machine with network interface"""
    async with ClientSecretCredential(
        cfg.tenant_id, cfg.client_id, cfg.client_secret
    ) as cred:
        cred = cast(AsyncTokenCredential, cred)  # fix library type errors
        async with NetworkManagementClient(cred, cfg.subscription_id) as nmc:
            async with ComputeManagementClient(cred, cfg.subscription_id) as cmc:
                return await delete_node(nmc, cmc, log, cfg, host)



"""Variables"""

from os import getenv

CONFIG_FILE = getenv("YASCHEDULER_CONF_PATH", "/etc/yascheduler/yascheduler.conf")
LOG_FILE = getenv("YASCHEDULER_LOG_PATH", "/var/log/yascheduler.log")
PID_FILE = getenv("YASCHEDULER_PID_PATH", "/var/run/yascheduler.pid")



"""Cloud API manager"""

import asyncio
import logging
from asyncio.locks import Lock
from pathlib import Path
from typing import Mapping, Optional, Sequence, Set, Union

from attrs import define, field
from typing_extensions import Self

from ..config import ConfigCloud, ConfigLocal, ConfigRemote, EngineRepository
from ..db import DB
from .adapters import azure_adapter, hetzner_adapter, upcloud_adapter
from .cloud_api import CloudAPI
from .protocols import CloudCapacity, PCloudAdapter, PCloudAPI, PCloudAPIManager


@define(frozen=True)
class CloudAPIManager(PCloudAPIManager):
    """Cloud API manager"""

    apis: Mapping[str, PCloudAPI] = field()
    db: DB = field()
    log: logging.Logger = field()
    on_tasks: Set[int] = field(init=False, factory=set)
    keys_dir: Path = field(factory=Path)
    allocation_lock: Lock = field(factory=Lock, init=False)

    @classmethod
    async def create(
        cls,
        db: DB,
        local_config: ConfigLocal,
        remote_config: ConfigRemote,
        cloud_configs: Sequence[ConfigCloud],
        engines: EngineRepository,
        log: Optional[logging.Logger] = None,
    ) -> Self:
        "Create cloud API manager"
        if log:
            log = log.getChild(cls.__name__)
        else:
            log = logging.getLogger(cls.__name__)

        adapters: Sequence[PCloudAdapter] = [
            azure_adapter,
            hetzner_adapter,
            upcloud_adapter,
        ]
        apis: Mapping[str, PCloudAPI] = {}

        def filter_adapters(prefix: str):
            return filter(lambda x: x.name == prefix, adapters)

        ssh_key_lock = asyncio.Lock()
        for cfg in cloud_configs:
            if cfg.max_nodes <= 0:
                log.debug("Cloud %s is skipped because of <1 max nodes", cfg.prefix)
                continue
            for adapter in filter_adapters(cfg.prefix):
                apis[adapter.name] = await CloudAPI.create(
                    adapter=adapter,
                    config=cfg,
                    local_config=local_config,
                    remote_config=remote_config,
                    engines=engines,
                    ssh_key_lock=ssh_key_lock,
                    log=log,
                )
        log.info("Active cloud APIs: %s", (", ".join(apis.keys()) or "-"))

        return cls(
            apis=apis,
            db=db,
            log=log,
            keys_dir=local_config.keys_dir,
        )

    def __bool__(self) -> bool:
        return bool(len(self.apis))

    async def stop(self) -> None:
        self.log.info("Stopping clouds...")

    def mark_task_done(self, on_task: int) -> None:
        self.on_tasks.discard(on_task)

    async def get_capacity(self) -> Mapping[str, CloudCapacity]:
        data = {}
        for name, count in (await self.db.count_nodes_clouds()).items():
            api = self.apis.get("name")
            data[name] = CloudCapacity(
                name=name,
                current=count,
                max=api.config.max_nodes if api else 0,
            )

        for api in self.apis.values():
            if api.name not in data:
                data[api.name] = CloudCapacity(
                    name=api.name, current=0, max=api.config.max_nodes
                )
        return data

    async def select_best_provider(
        self, want_platforms: Optional[Sequence[str]] = None
    ) -> Optional[PCloudAPI]:
        """Select best cloud API"""
        self.log.debug("Enabled providers: %s", ", ".join(self.apis.keys()))
        used_providers = []
        suitable_providers = list(self.apis.keys())

        cap = await self.get_capacity()

        for name, capacity in cap.items():
            used_providers.append((name, capacity.current))
            api = self.apis.get(name)
            if not api:
                continue
            # remove maxed out providers
            if capacity.current >= api.config.max_nodes:
                suitable_providers.remove(api.name)
                continue
            # remove not supported platforms
            if want_platforms:
                if not any(map(api.is_platform_supported, want_platforms)):
                    suitable_providers.remove(api.name)

        self.log.debug("Used providers: %s", used_providers)
        if not suitable_providers:
            self.log.debug("No suitable cloud providers")
            return

        ok_apis = filter(lambda x: x.name in suitable_providers, self.apis.values())
        ok_apis_sorted = sorted(ok_apis, key=lambda x: x.config.priority, reverse=True)
        api = ok_apis_sorted[0]
        self.log.debug("Chosen: %s", api.name)
        return api

    async def allocate_node(
        self, want_platforms: Optional[Sequence[str]] = None, throttle: bool = False
    ):
        """Allocate new node"""
        async with self.allocation_lock:
            api = await self.select_best_provider(want_platforms)
            if not api:
                return
            if throttle and api.get_op_semaphore().locked():
                self.log.debug(f"Cloud {api.name} is overloaded by requests")
                await asyncio.sleep(1)
                return

            tmp_ip = await self.db.add_tmp_node(api.name, api.config.username)
            await self.db.commit()
        try:
            ip_addr = await api.create_node()
        finally:
            await self.db.remove_node(tmp_ip)
            await self.db.commit()

        await self.db.add_node(ip_addr, api.config.username, None, api.name, True)
        await self.db.commit()
        return ip_addr

    async def allocate(
        self,
        on_task: Optional[int] = None,
        want_platforms: Optional[Sequence[str]] = None,
        throttle: bool = True,
    ) -> Union[str, None]:
        if on_task in self.on_tasks:
            return
        if on_task:
            self.on_tasks.add(on_task)
        try:
            return await self.allocate_node(want_platforms, throttle)
        except Exception as err:
            self.log.error(f"Can't allocate node: {err}")
            if on_task:
                self.mark_task_done(on_task)
        return

    async def deallocate(self, ip_addr: str):
        node = await self.db.get_node(ip_addr)
        if not node or not node.cloud:
            return
        if node.cloud not in self.apis:
            self.log.warning(
                f"Can't deallocate node {node.ip} - unsupported cloud {node.cloud}"
            )
        await self.db.disable_node(ip_addr)
        await self.db.commit()
        try:
            await self.apis[node.cloud].delete_node(node.ip)
        except Exception as err:
            self.log.error(f"Can't deallocate node {node.ip}: {err}")
            return False
        await self.db.remove_node(node.ip)
        await self.db.commit()



"""
Aiida plugin for yascheduler,
with respect to the supported yascheduler engines
"""

import aiida.schedulers  # pylint: disable=import-error
from aiida.orm import load_node  # pylint: disable=import-error

# pylint: disable=import-error
from aiida.schedulers.datastructures import JobInfo, JobState, NodeNumberJobResource

_MAP_STATUS_YASCHEDULER = {
    "TO_DO": JobState.QUEUED,
    "RUNNING": JobState.RUNNING,
    "DONE": JobState.DONE,
}
_CMD_PREFIX = ""  # NB under virtualenv, this should refer to virtualenv's /bin/


class YaschedJobResource(NodeNumberJobResource):
    def __init__(self, *_, **kwargs):
        super().__init__(**kwargs)


class YaScheduler(aiida.schedulers.Scheduler):
    """
    Support for the YaScheduler designed specifically for MPDS
    """

    _logger = aiida.schedulers.Scheduler._logger.getChild("yascheduler")

    # Query only by list of jobs and not by user
    _features = {
        "can_query_by_user": False,
    }

    # The class to be used for the job resource.
    _job_resource_class = YaschedJobResource

    def _get_joblist_command(self, jobs=None, user=None):
        """
        The command to report full information on existing jobs.
        """

        # pylint: disable=import-error
        from aiida.common.exceptions import FeatureNotAvailable

        if user:
            raise FeatureNotAvailable("Cannot query by user in Yascheduler")
        command = [f"{_CMD_PREFIX}yastatus"]
        # make list from job ids (taken from slurm scheduler)
        if jobs:
            joblist = []
            if isinstance(jobs, str):
                joblist.append(jobs)
            else:
                if not isinstance(jobs, (tuple, list)):
                    raise TypeError(
                        "If provided, the 'jobs' variable must be a string or a list of strings"
                    )
                joblist = jobs
            command.append("--jobs {}".format(" ".join(joblist)))
        return " ".join(command)

    def _get_detailed_jobinfo_command(self, jobid):
        """
        Return the command to run to get the detailed information on a job,
        even after the job has finished.
        """
        return f"{_CMD_PREFIX}yastatus --jobs {jobid}"

    def _get_submit_script_header(self, job_tmpl):
        """
        Return the submit script header, using the parameters from the
        job_tmpl.
        """
        assert job_tmpl.job_name
        # There is no other way to get the code label and the WF uuid except this (TODO?)
        pk = int(job_tmpl.job_name.split("-")[1])
        aiida_node = load_node(pk)

        # We map the lowercase code labels onto yascheduler engines,
        # so that the required input file(s) can be deduced
        lines = [f"ENGINE={aiida_node.inputs.code.label.lower()}"]

        try:
            lines.append(f"PARENT={aiida_node.caller.uuid}")
        except AttributeError:
            pass

        lines.append(f"LABEL={job_tmpl.job_name}")
        return "\n".join(lines)

    def _get_submit_command(self, submit_script):
        """
        Return the string to execute to submit a given script.
        """
        return f"{_CMD_PREFIX}yasubmit {submit_script}"

    def _parse_submit_output(self, retval, stdout, stderr):
        """
        Parse the output of the submit command, as returned by executing the
        command returned by _get_submit_command command.
        """
        if stderr.strip():
            self.logger.warning(f"Stderr when submitting: {stderr.strip()}")

        output = stdout.strip()

        try:
            int(output)
        except ValueError:
            self.logger.error("Submitting failed, no task id received")

        return output

    def _parse_joblist_output(self, retval, stdout, stderr):
        """
        Parse the queue output string, as returned by executing the
        command returned by _get_joblist_command command,
        that is here implemented as a list of lines, one for each
        job, with _field_separator as separator. The order is described
        in the _get_joblist_command function.
        Return a list of JobInfo objects, one of each job,
        each relevant parameters implemented.
        """
        if stderr.strip():
            self.logger.warning(f"Stderr when parsing joblist: {stderr.strip()}")
        job_list = [job.split() for job in stdout.split("\n") if job]
        job_infos = []
        for job_id, status in job_list:
            job = JobInfo()
            job.job_id = job_id
            job.job_state = _MAP_STATUS_YASCHEDULER[status]
            job_infos.append(job)
        return job_infos

    def _get_kill_command(self, jobid):
        """
        Return the command to kill the job with specified jobid.
        """

    def _parse_kill_output(self, retval, stdout, stderr):
        """
        Parse the output of the kill command.
        """



#!/usr/bin/env python3

from subprocess import DEVNULL
from typing import Optional

from asyncssh.connection import SSHClientConnection
from asyncssh.process import SSHClientProcess, SSHCompletedProcess
from attrs import define

from .protocol import QuoteCallable


@define(frozen=True)
class ProcessInfo:
    pid: int
    name: str
    command: str


async def run(
    conn: SSHClientConnection,
    quote: QuoteCallable,
    command: str,
    *args,
    cwd: Optional[str] = None,
    **kwargs,
) -> SSHCompletedProcess:
    """
    Run process and wait for exit
    :raises asyncssh.Error: An SSH error has occurred.
    """
    if cwd:
        command = f"cd {quote(cwd)}; {command}"
    return await conn.run(command, *args, **kwargs)


async def run_bg(
    conn: SSHClientConnection,
    quote: QuoteCallable,
    command: str,
    *args,
    cwd: Optional[str] = None,
    **kwargs,
) -> SSHClientProcess:
    """
    Create background process.
    :raises asyncssh.ChannelOpenError: An SSH error has occurred.
    """
    if cwd:
        command = f"cd {quote(cwd)}; {command}"
    return await conn.create_process(
        command, *args, **kwargs, stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL
    )



"""Hetzner cloud methods"""

import asyncio
import logging
from concurrent.futures.thread import ThreadPoolExecutor
from functools import lru_cache, partial
from typing import Optional

from asyncssh.public_key import SSHKey as ASSHKey
from hcloud import APIException
from hcloud import Client as HClient
from hcloud.images.domain import Image
from hcloud.server_types.domain import ServerType
from hcloud.servers.client import BoundServer
from hcloud.ssh_keys.domain import SSHKey as HSSHKey

from ..config import ConfigCloudHetzner
from .protocols import PCloudConfig
from .utils import get_key_name, get_rnd_name

executor = ThreadPoolExecutor(max_workers=5)


@lru_cache(maxsize=None)
def get_client(cfg: ConfigCloudHetzner) -> HClient:
    "Get Hetzner client"
    return HClient(cfg.token)


@lru_cache
def get_ssh_key_id(client: HClient, key: ASSHKey) -> int:
    "Get Hetzner ssh id"
    key_name = get_key_name(key)
    pub_key = key.export_public_key("openssh").decode("utf-8")

    try:
        return client.ssh_keys.create(name=key_name, public_key=pub_key).id
    except APIException as err:
        if "already" in str(err):
            hkey = client.ssh_keys.get_by_fingerprint(
                key.get_fingerprint("md5").split(":", maxsplit=1)[1]
            ) or client.ssh_keys.get_by_name(key_name)
            if hkey:
                return hkey.id
            prefix = "yakey"
            name_len = len(get_rnd_name(prefix))
            for hkey in client.ssh_keys.get_all():
                if hkey.name.startswith(prefix) and len(hkey.name) == name_len:
                    return hkey.id
        raise err


async def hetzner_create_node(
    log: logging.Logger,
    cfg: ConfigCloudHetzner,
    key: ASSHKey,
    cloud_config: Optional[PCloudConfig] = None,
) -> str:
    """Create node"""
    loop = asyncio.get_running_loop()
    client = await loop.run_in_executor(executor, get_client, cfg)
    ssh_key_id = await loop.run_in_executor(executor, get_ssh_key_id, client, key)

    create_server = partial(
        client.servers.create,
        name=get_rnd_name("node"),
        server_type=ServerType(cfg.server_type),
        image=Image(name=cfg.image_name),
        ssh_keys=[HSSHKey(id=ssh_key_id, name=get_key_name(key))],
        user_data=cloud_config.render() if cloud_config else None,
    )
    response = await loop.run_in_executor(executor, create_server)
    server = response.server
    ip_addr = server.public_net.ipv4.ip
    log.info("CREATED %s", ip_addr)
    return ip_addr


def find_srv(client: HClient, host: str) -> Optional[BoundServer]:
    """Find BoundServer by IP addr"""
    for server in client.servers.get_all():
        if server.public_net.ipv4.ip == host:
            return client.servers.get_by_id(server.id)
    return None


async def hetzner_delete_node(
    log: logging.Logger,
    cfg: ConfigCloudHetzner,
    host: str,
):
    """Delete node"""
    loop = asyncio.get_running_loop()
    client = await loop.run_in_executor(executor, get_client, cfg)
    server = await loop.run_in_executor(executor, find_srv, client, host)

    if server:
        await loop.run_in_executor(executor, server.delete)
        log.info("DELETED %s", host)

    else:
        log.info("NODE %s NOT DELETED AS UNKNOWN", host)
